; ====================================================================================
; UT5 Monitor ROM assembler source code for RCA CDP18S020 development board
;
; This is a regenerated assembler source code image for the UT5 (Microterminal)
; monitor ROM for the RCA CDP18S020 development board. This was originally supplied
; with the CDP18S021 Microterminal (marked CDPR522), and replaces the UT4 serial 
; monitor in socket U2 on the board. This code will assemble to produce an exact
; image of the original monitor image, except for one undefined byte (set to zero here).
;
; This image has been generated by hand editing the hex file for the UT5 monitor from
; the listing in the MPM-212 RCA COSMAC Microterminal manual, then using the DASMX
; disassembler to generate a (poor quality) assembler file. After much touching up
; and redefinition of labels, etc., the excellent a18.exe assembler has been used to
; regenerate a new hex file that matches the original one. The assembler source has 
; also been compared to the original MPM-212 listing to verify its integrity.
;
; The source code uses some weird and wonderful tricks within the 1802 CPU, including
; the SEP4 and SEP5 call/return mechanism, and the more subtle next byte OUT mechanism.
; While the a18 assembler has an elegant set of pseudo-ops to make the call/return
; mechanism more readable, the OUT mechanism results in a db data definition appearing
; within the source code - even the original MPM-212 listing included this.
;
; This listing has been refined to be fully relocatable, so should be adaptable for other
; projects where a Microterminal is being interfaced to a 1802-based system.
; Note that there are some expedient register setups, which may break if the code is
; arbitrarily relocated; these are marked with ###.
;
; ====================================================================================

; Ensure we are building this for a 1802 CPU without any enhancements
        CPU     1802

; ====================================================================================
; Useful definitions for this module

; -------------------------------------------------------------------------------------
; Base register definitions for this module
; Where registers have a specific use in the monitor, this is indicated in comments
; NOTE: Registers have different usages in the separate user-accessible LEDD routine

r0	        EQU	0           ; only used to start execution
r1		EQU	1           ; not used anywhere
r2		EQU	2           ; used for memory read/write
r3		EQU	3           ; used as program counter
r4		EQU	4           ; used in call/return mechanism
r5		EQU	5           ; used in call/return mechanism
r6		EQU	6           ; used in call/return mechanism
r7		EQU	7           ;
r8		EQU	8           ; only used for delays
r9		EQU	9           ; not used anywhere
r10		EQU	10          ;
r11		EQU	11          ;
r12		EQU	12          ; bit 0 used to select address/data entry mode
r13		EQU	13          ; mostly used or display lookup
r14		EQU	14          ; mostly used for selected data value (8 bits)
r15		EQU	15          ; mostly used for selected address (16 bits)

; -------------------------------------------------------------------------------------
; Fundamental definitions for this module

; stack is set to the top of the special 32 byte CDP1824 register storage RAM 
STACK_TOP	EQU	$8C1F

; LED digit storage is set to bottom of special 32 byte CDP1824 register storage RAM
DIGITS          EQU     $8C00

; user program return address after calling ENTRY routine
ENTRY_RET       EQU     $0005

; ====================================================================================
; Define starting point for this ROM image in the target system
	org	$8000

; ====================================================================================
; Main entry point for the UT5 monitor program - executed whenever board is reset
; This refreshes the display and scans for keyboard input, performing actions as required.

KEYUT:
; Disable interrupts and begin program execution, register r0 is PC and X at this point
	dis                         ; disable interrupts and return to r0
	db	$00                 ; defines r0 for both X and P on return
; Perform dummy data fetches to assist CPU register readout function
; These put all of the register addresses on the address bus
	ldn	r2
	ldn	r3
	ldn	r4
	ldn	r5
	ldn	r6
	ldn	r7
	ldn	r8
	ldn	r9
	ldn	r10
	ldn	r11
	ldn	r12
	ldn	r13
	ldn	r14
	ldn	r15
	ldn	r1
; Store pointer to OPTION subroutine in r3 register
; This routine clears the address to zero and starts normal keyboard/display scan
	ldi	HIGH(OPTION)
	phi	r3
	ldi	LOW(OPTION)
	plo	r3
; jump to the initialisation code, which will progress to OPTION afterwards	
        lbr	INIT

; ....................................................................................
; The main LED display driver subroutine
; Scans display with 6 digits of info, digits 5 and 6 blank
; On entry, R12 contains operating mode (address or data)
LED:
; set starting digit logical address in R7 for accessing RAM
	ldi	LOW(DIGITS)
	plo	r7
; set high byte of display segment lookup table using call address (hacky!)
; this should be adjusted if we wish to relocate the code in the future ###
	ghi	r4
	phi	r13
; set starting digit selection code in R11, and setup R2 for IO addressing
	ldi	$80
	plo	r11
	sex	r2                  ; use R2 for IO address selection
LOOP1:
; digit must be displayed, so obtain two free RAM bytes from stack 
	dec	r2
	dec	r2
; transfer digit selection code into RAM, ready for use
	glo	r11
	str	r2
; obtain display data, and use display table to look up 7 segment data
	lda	r7                  ; load data and advance digit address
	ani	$0F
	adi	LOW(DTAB)
	plo	r13
; obtain converted data from table, and store result in R11 high
	lda	r13                 ; load data from table
	phi	r11
; now output digit selection code from RAM addressed by R2 (auto increment)
	out	4
; get address/data operating mode from R12 low and check for selection
	glo	r12                 ; get operating mode
	ani	$01                 ; check if address (set) or data (clear)
	bz	DISD                ; jump if operating in data mode
; operating in address mode, so check if address digits are selected
	glo	r11
	ani	$0F
	bnz	OUTD                ; not address mode, so jump
LDOT:
; illuminate the decimal point in the 7 segment display
	ghi	r11
	ani	$BF
	br	DELAY               ; jump to display digit
DISD:
; operating in data mode, so check for data digits selected
	glo	r11
	ani	$FC
	bz	LDOT                ; illuminate decimal point if so
OUTD:
; obtain display data from R11, ready for outputting
	ghi	r11
DELAY:
; now store display data in RAM and output to display segments
	str	r2
	out	3
; run delay subroutine via call/return mechanism
	sep	r4
        dw      DELSUB
; Turn off the segment via a weird and wonderful construct.
; r3 is the current PC, so when used for the output, address is the
; memory location following the OUT instruction
	sex	r3                  ; set X register to r3 (PC) for I/O
	out	3                   ; write next byte to the port, N=3
        db      $FF                 ; data to write to output port
	sex     r2                  ; restore X register to R2
CHEND:
; check if we have reached the end of the scan (remember only 6 digits active)
	glo	r7                  ; obtain logical digit address code
	smi	LOW(DIGITS+$06)     ; check if reached end digit address
	bnf	CDLB                ; jump if not reached the end
; finished scan, so exit from the routine
	sep	r5

CDLB:
; shift down the digit select, and check for the two blank digits
	glo	r11
	shr
	plo	r11
; check if selecting the two blanked digits
	ani	$F3                 ; digits 5 and 6 will be blanked
	bnz	LOOP1               ; loop round for non-blank digit
; blank digit, so simply jump back to end check for another cycle
; we could just jump back to local shift down, as end check will not fire!
	br	CHEND

; ....................................................................................
; Main monitor keyboard scan and display update routine
HKEY1:
; scan the display and check for key being pressed (ext 3 flag SET)
; scan and refresh LED display via call/return mechanism
	sep	r4
	dw	LED
	b3	HKEY1               ; loop back if key still pressed
HKEY2:
; key has now been released, so perform debounce delay
; scan and refresh LED display via call/return mechanism
	sep	r4
	dw	LED
; get address/data operating mode from R12 low and check for selection
	glo	r12
	ani	$01                 ; check if address (set) or data (clear)
	bz	HKEY2B              ; jump if in data selection mode
; address select mode, so update LED data display via call/return mechanism
	sep	r4
        dw      UPDRE0              ; fetch data from memory and store in buffer
HKEY2B:
; loop round while the key remains released (while ext 3 flag CLEAR)
	bn3	HKEY2               ; loop round if key still released
; key has now been pressed, so perform debounce delay
; scan and refresh LED display via call/return mechanism
	sep	r4
	dw	LED
; set high byte of display segment lookup table using call address (hacky!)
; this should be adjusted if we wish to relocate the code in the future ###
	ghi	r4
	phi	r10
; preset keyboard lookup table address low byte, ready for lookup
	ldi	LOW(KTAB)
	plo	r10
; read in the encoded keyboard value from the port into R2 addressed memory
	inp	4
FCOM:
; load value from table into R10, auto incrementing, and match with input data
	lda	r10                 ; load data from table, increment addr
	xor                         ; match to read in keyboard data
	bz	FMAT                ; jump if match with input key
; no match, so check if we have reached the end of the keyboard table
; if we reach the end, key not recognised, so reset to the start
	glo	r10                 ; get table address
	smi	LOW(KTAB+$14)       ; check if hit end of table
	bdf	HKEY1               ; finished scan, reset to main keyboard handler
; try again with the next entry from the keyboard table
	br	FCOM                
FMAT:
; check if this is a hex key that has been pressed, by removing hex offset
; the lookup table address reflects the hex key code that has been loaded
	glo	r10                 ; obtain table address
	smi	LOW(KTAB+$01)       ; remove incremented offset
	plo	r10                 ; store back in register
; check if the result is in hex digit range
	ani	$F0                 ; mask out non-hex values
	bnz	SPBRN               ; not a hex digit, so branch to operation
; hex value, so shift data into display via call/return mechanism
	sep	r4
        dw      SHIFTD
; loop back round to main keyboard scan routine
	br	HKEY1

; special function key handler routine
SPBRN:
; mask off top bits of returned table index and look up in function table
	glo	r10
	ani	$0F
; use this value to select low byte lookup in function table
	adi	LOW(BTAB)
	plo	r10
; set high byte of function lookup table using PC address (hacky!)
; this should be adjusted if we wish to relocate the code in the future ###
	ghi	r3
	phi	r10
; load in function table address and store in PC - will automatically branch
	lda	r10
	plo	r3

; Special function key selection lookup table (indexed by decoded key code)
; Only four function keys are scanned as part of the matrix
; NOTE: If code is updated/relocated, we may need to adjust things here ###
; All these entry points must appear within the same page!
BTAB:
        db      LOW(XCHN)           ; Exchange data and address mode (<-> key)
        db      LOW(DOLP)           ; DOLP - start addressed program ($P key)
        db      LOW(INC)            ; INC - increment current address (INC key)
        db      LOW(OPTION)         ; OPTION - clear address display (CA key) 

; ....................................................................................
SHIFTD:
; shifts new digit into the display, either address or data
	glo	r12                 ; get operating mode
	ani	$01                 ; check if address (set) or data (clear)
	bz	DAS                 ; jump if operating in data entry mode
ADS:
; address entry mode, so shift into address display - r10 contains the hex key value
; first shift up the top byte, losing existing top nibble
	ghi	r15                 ; obtain top addr byte
	shl                         ; shift up to top nibble, bottom nibble zero
	shl                         ;
	shl                         ;
	shl                         ;
	str	r2                  ; temporary store in memory
; now shift top lower nibble into bottom upper nibble
	glo	r15                 ; obtain bottom addr byte
	shr                         ; shift down to bottom nibble, top nibble zero
	shr                         ;
	shr                         ;
	shr                         ;
	or                          ; mask in previous shifted value
	phi	r15                 ; store back in top addr byte
; now shift bottom lower nibble into top lower nibble and store
	glo	r15                 ; obtain bottom addr byte
	shl                         ; shift up to top nibble, bottom nibble zero
	shl                         ;
	shl                         ;
	shl                         ;
	str	r2                  ; temporary store in memory
	glo	r10                 ; obtain hex key value
	or                          ; OR into the display data
	plo	r15                 ; and store in addr bottom byte
; continue to generic routine, r15 contains updated address display value

UPDRF:
; now update the address display with the updated contents of r15
; We convert 16 bit value into four separate hex digits in store
	ldi	LOW(DIGITS)         ; select first address display digit
	plo	r7                  ; address in r7
; update the first digit from top nibble of top r15 address byte
	ghi	r15                 ; obtain top address byte
	shr                         ; shift down to lower nibble, top nibble zero
	shr                         ;
	shr                         ;
	shr                         ;
	str	r7                  ; store in first digit pos
; update the second digit from bottom nibble of top r15 byte
	inc	r7                  ; select next digit buffer
	ghi	r15                 ; obtain top address byte
	ani	$0F                 ; mask out top nibble
	str	r7                  ; store in second digit pos
; update the third digit from top nibble of bottom r15 byte
	inc	r7                  ; select next digit buffer
	glo	r15                 ; obtain bottom address byte
	shr                         ; shift down to lower nibble, top nibble zero
	shr                         ;
	shr                         ;
	shr                         ;
	str	r7                  ; store in third digit pos
; update the fourth digit from bottom nibble of bottom r15 byte
	inc	r7                  ; select next digit buffer
	glo	r15                 ; obtain bottom address byte
	ani	$0F                 ; mask out top nibble
	str	r7                  ; store in fourth digit pos

UPDRE0:
; fetch new data from memory location and store in buffer
; we do this because the selected address will have changed by user input
	lda	r15                 ; load from memory address (and increment)
	plo	r14                 ; store in data register
	dec	r15                 ; adjust address back (remove increment)

UPDBY:
; generic routine for updating data display memory
; We convert 16 bit value into two separate hex digits in store
	ldi	LOW(DIGITS+$04)     ; point to fifth digit (first data digit)
	plo	r7                  ; address in r7
; store fifth digit in buffer, from top nibble of display data
	glo	r14                 ; obtain the data value byte
	shr                         ; shift down to lower nibble, top nibble zero
	shr                         ;
	shr                         ;
	shr                         ;
	str	r7                  ; store in fifth digit pos
; store sixth digit in buffer, from bottom nibble of display data
	inc	r7                  ; select next digit buffer
	glo	r14                 ; obtain the data value byte
	ani	$0F                 ; mask out top nibble
	str	r7                  ; store in sixth digit pos
; display data formatted, so exit from routine
	sep	r5

DAS:
; data entry mode, so shift into data display - r10 contains the hex key value
	glo	r14                 ; obtain data byte value
	shl                         ; shift up to top nibble, bottom nibble zero
	shl                         ;
	shl                         ;
	shl                         ;
	str	r2                  ; temporary store in memory
	glo	r10                 ; obtain hex key value
	or                          ; OR into the display data
	plo	r14                 ; and store in data display
; jump to the generic data update routine, r14 contains updated data display value
	br	UPDBY

; ....................................................................................
; Various function key handler routines, accessed by BTAB table

DOLP:
; Start execution at current address ($P key pressed)
; Note that stack is also set here, so the callee must sort out its own stack!
; First transfer address register into the r0 register
	ghi	r15
	phi	r0
	glo	r15
	plo	r0
; Now set X register and then program counter to this address
	sex	r0                  ; set stack/IO to spec address
	sep	r0                  ; transfers execution to selected address

INC:
; Store data and increment address (INC key pressed)
; Always stores data in current address before incrementing address
	glo	r14                 ; obtain current data
	str	r15                 ; store in current address
	inc	r15                 ; increment address
; continue to common code to update display and scan keyboard
	skp			    ; skip to address SKIP1

XCHN:
; Exchange address/data selection mode (<-> key pressed)
; Simply increment mode register to toggle bit 0
	inc	r12
SKIP1:
; continue to common code to update display and scan keyboard
	lbr	CUP

OPTION:
; Clear Address option (CA key pressed), so reset address counter to zero
; This is also the main entry point after system initialisation!
	ldi	$00
	phi	r15
	plo	r15
CUP:
; common entry point to update display via call/return mechanism
	sep	r4
        dw      UPDRF
; return to main key handler routine
	lbr	HKEY1

; ....................................................................................
; Special entry point for call/return technique initialisation
; R2 = stack pointer (default $8C1F top of stack)
; R4 = call subroutine pointer
; R5 = return subroutine pointer
; Eventually returns execution to location $0005
; This is called by the user to setup call/return mechanism
ENTRY:
; set up return entry point in user program space
	ldi	HIGH(ENTRY_RET)
	phi	r3
	ldi	LOW(ENTRY_RET)
	plo	r3                  

; Main initialisation code before starting normal execution
; On entry r3 points to a specific return routine, set by the caller
INIT:
; store address of call subroutine in R4
	ldi	HIGH(CALL)
	phi	r4
	phi	r5                  ; for expediency, set same high address in R5
	ldi	LOW(CALL)
	plo	r4
; store address of return subroutine in R5 
	ldi	LOW(RETPGM)
	plo	r5
; store defined stack top location in R2
	ldi	HIGH(STACK_TOP)
	phi	r2
	phi	r7                  ; for expediency, set same high address in R7
	ldi	LOW(STACK_TOP)
	plo	r2
; initialise operating mode in an expedient manner (bit 0 is set, address entry)
; should STACK_TOP change to an even address, this will adversely affect mode ###
	plo	r12                 ; for expediency, set same low address in R12
; now set X register (stack access) to r2, and program counter to r3
; The r3 register points to the routine for subsequent execution
	sex	r2                  ; set X to access stack
	sep	r3                  ; transfers execution to r3 address

; ....................................................................................
; Keyboard lookup table follows - first 16 entries map to key hex values 
; last four map to selected function keys (<->, $P, INC and CA in order)
KTAB:
	db	$0A, $8E, $80, $81      ; 0, 1, 2, 3
	db	$4E, $40, $41, $2E      ; 4, 5, 6, 7
	db	$20, $21, $1E, $10      ; 8, 9, A, B
	db	$11, $06, $08, $04      ; C, D, E, F
	db	$02, $09, $01, $0C      ; <->, $P, INC, CA

; Display lookup table follows - converts hex to LED segment definitions
; last entry completely blanks the display digit
DTAB:
	db	$44, $7D, $C1, $51      ; 0, 1, 2, 3
	db	$78, $52, $42, $5D      ; 4, 5, 6, 7
	db	$40, $50, $48, $62      ; 8, 9, A, B
	db	$C6, $61, $C2, $CA      ; C, D, E, F
	db	$FF                     ; blank display

; ....................................................................................
; Delay subroutine for timing segment illumination; though the Microterminal has
; a built-in hardware delay for protection, we still use a software delay for this,
; and this will effectively control the brightness of the display.
;
DELSUB:
; set the initial delay value in r13
	ldi	$20
	plo	r13
DL:
; decrement counter, and loop round until zero
	dec	r13
	glo	r13
	bnz	DL
; delay finished, so return to called
	sep	r5

; ....................................................................................
; Independent routine which lists memory sequentially from $0000 onwards, waiting
; for approximately 1 second between updates. From how it appears, it will actually
; start displaying from $0001, because it immediately increments before the first
; update to the LED display.

COUNT:
; Specify the INCR routine as the return entry point, and run the initialiser
	ldi	HIGH(INCR)
	phi	r3
	ldi	LOW(INCR)
	plo	r3
	lbr	INIT                ; execute initialisation routine

; Entry point for the auto incrementer display
INCR:
; set the starting address to zero for the initial display
; however, it is immediately incremented, so the display actually runs from $0001
	ldi	$00
	phi	r15
	plo	r15

INCRM:
; increment the address, ready for next display
	inc	r15
; set up the delay loop counter before updating the LED display
	ldi	$80                 ; set delay counter value
	plo	r8
; update the LED display address/data via call/return mechanism
	sep	r4
        dw      UPDRF
SHOW:
; scan LED display via call/return mechanism
	sep	r4
        dw      LED
; now decrement the delay counter and loop round until done
; this means the display will be continually scanned during the delay
	dec	r8
	glo	r8
	bnz	SHOW
; finished delay, so loop round to increment memory
	br	INCRM

; ....................................................................................
; Generic LED display update routine. Shows contents of display memory
; addressed by register R15 on the LED display. Memory data should contain
; hex data in bottom nibble, with high nibble containing control data (zero
; for normal display, $1 for DP on, $8 to $F to blank digit.
; This is for use by user programs, and does not run as part of the monitor.
LEDD:
; start with left most digit in display (digit selection code)
	ldi	$80
	plo	r14
; set high byte of display segment lookup table
	ldi	HIGH(DTAB)
	phi	r13
LOOPD:
; check if top bit of data is set, indicating blank digit
	ldn	r15
	ani	$80                 ; 
	bnz	SKIPD               ; skip digit if blank
; digit must be displayed, so obtain two free RAM bytes from stack 
	dec	r2
	dec	r2
; transfer digit selection code into RAM, ready for use
	glo	r14
	str	r2
; obtain display data, and use display table to look up 7 segment data
	ldn	r15
	ani	$0F
	adi	LOW(DTAB)
	plo	r13
; obtain converted data from table, and store result in R14 high
	lda	r13
	phi	r14
; now check if decimal point must be lit for this digit
	ldn	r15
	ani	$10
	bz	DISPD               ; skip if DP not lit
; mask in the DP lit into the 7 segment data
	ghi	r14
	ani	$BF
	phi	r14
DISPD:
; now output digit selection code from RAM addressed by R2 (auto increment)
	out	4
; now obtain display data, store in RAM and output to display segments
	ghi	r14
	str	r2
	out	3
; run delay subroutine via call/return mechanism
	sep	r4
        dw      DELSUB
; Turn off the segments via a weird and wonderful construct.
; r3 is the current PC, so when used for the output, address is the
; memory location following the OUT instruction
	sex	r3                  ; set X register to r3 (PC) for I/O
	out	3                   ; write next byte to the port, N=3
        db      $FF                 ; data to write to output port
	sex     r2                  ; restore X register to R2
SKIPD:
; increment to the next digit in memory, and adjust digit selection code
	inc	r15
	glo	r14
	shr
	plo	r14
; loop round until all digits have been updated
	bnz	LOOPD
; reset specified register address back to the starting point (rather crude!)
	dec	r15
	dec	r15
	dec	r15
	dec	r15
	dec	r15
	dec	r15
	dec	r15
	dec	r15
; return back to the caller
	sep	r5

; ....................................................................................
; Displays registers R10 (left) and R11 (right) on the LED display
; Data in both registers is standard hex data, which will be converted
REGDIS:
; set LED display starting address in register R15
	ldi	HIGH(DIGITS)
	phi	r15
	ldi	LOW(DIGITS)
	plo	r15
; format R10 contents onto the left hand LED display digits
; format data into digit 1
	ghi	r10                 ; obtain high byte of r10
	shr                         ; shift top nibble to bottom
	shr
	shr
	shr
	str	r15                 ; store in digit address
; format data into digit 2
	inc	r15                 ; select next digit
	ghi	r10                 ; obtain high byte of r10
	ani	$0F                 ; mask out top nibble
	str	r15                 ; store in digit address
; format data into digit 3
	inc	r15                 ; select next digit
	glo	r10                 ; obtain low byte of r10
	shr                         ; shift top nibble to bottom
	shr
	shr
	shr
	str	r15                 ; store in digit address
; format data into digit 4
	inc	r15                 ; select next digit
	glo	r10                 ; obtain low byte of r10
	ani	$0F                 ; mask out top nibble
	str	r15                 ; store in digit address
; format R11 contents onto the right hand LED display digits
; format data into digit 5
	inc	r15                 ; select next digit
	ghi	r11                 ; obtain high byte of r11
	shr                         ; shift top nibble to bottom
	shr
	shr
	shr
	str	r15                 ; store in digit address
; format data into digit 6
	inc	r15                 ; select next digit
	ghi	r11                 ; obtain high byte of r11
	ani	$0F                 ; mask out top nibble
	str	r15                 ; store in digit address
; format data into digit 7
	inc	r15                 ; select next digit
	glo	r11                 ; obtain low byte of r11
	shr                         ; shift top nibble to bottom
	shr
	shr
	shr
	str	r15                 ; store in digit address
; format data into digit 8
	inc	r15                 ; select next digit
	glo	r11                 ; obtain low byte of r11
	ani	$0F                 ; mask out top nibble
	str	r15                 ; store in digit address
; set LED display starting address in R15 for generic LED display
	ldi	LOW(DIGITS)
	plo	r15
; call general purpose LED update via call/return mechanism
	sep	r4
        dw      LEDD
; return to the caller
	sep	r5

; ....................................................................................
; skip a byte (value undefined in original monitor listing)
	db	$00
; Specialised call/return mechanism for facilitating subroutine calls both within
; the monitor program and also for user routines. Usually, r4 should be pre-loaded with
; the address of the CALL routine, and RETPGM with the address of the RETPGM routine.
; When making the call, you do a sep r4, followed by the two subroutine address bytes.
; When making the return, you just do a set r5; execution will recommence at the 
; instruction immediately after the two subroutine bytes.
; The r6 register is used to hold return address, so should NOT be modified by subroutine

EXITA:
; Exit point for the call routine, using r3 as program counter
; This will jump to the required subroutine
	sep	r3                  ; jump to r3 address
CALL:
; Specialised call routine, used for calling subroutine address
; on entry, r3 will point to subroutine address, on exit, X register is set to r2
	sex	r2                  ; point to the stack
; preserve the r6 register on the stack
	ghi	r6                  ; save r6 high into the stack
	stxd                        ;
	glo	r6                  ; save r6 low into the stack
	stxd                        ;
; now transfer r3 register into r6 register (r3 points to subroutine address)
; this will be used to store the return address for later use
	ghi	r3                  ; transfer r3 high into r6 high
	phi	r6                  ;
	glo	r3                  ; transfer r3 low into r6 low
	plo	r6                  ;
; now load subroutine address into the r3 register, ready for execution
	lda	r6                  ; load memory and advance
	phi	r3                  ; store in r3 high
	lda	r6                  ; load memory and advance
	plo	r3                  ; store in r3 low
; the r6 register has advanced, and will therefore point the return address
; branch to exit point, which will execute the subroutine
	br	EXITA

EXITP:
; Exit point for the return routine, using r3 as program counter
; This will jump back to the calling program following subroutine address
	sep	r3                  ; jump to r3 address (caller)
RETPGM:
; Specialised return routine, used for returning from subroutine
; Transfer stored return address back into r3 for later use
	ghi	r6                  ; transfer r6 high into r3 high
	phi	r3                  ;
	glo	r6                  ; transfer r6 low into r3 low
	plo	r3                  ;
; now point to stack and recover the original value of r6
	sex	r2                  ; point to stack
	inc	r2                  ; increment to next byte
	ldxa                        ; load from stack and advance
	plo	r6                  ; store in r6 reg low
	ldx                         ; load from stack
	phi	r6                  ; store in r6 reg
; branch to exit point, which will return to the caller (after subroutine address)
	br	EXITP

; ====================================================================================
; end of monitor program listing assembly
    END

; ====================================================================================
; ====================================================================================
